# -*- coding: utf-8 -*-
# Modified work Copyright (c) 2017-2019 Science and Technology
# Facilities Council.
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Base classes for all Fortran statement types
"""

__all__ = ['Statement', 'BeginStatement', 'EndStatement', 'Variable',
           'AttributeHolder', 'ProgramBlock']

import copy
import logging

from six import with_metaclass
from fparser.common.readfortran import Line, Comment
from fparser.common.utils import split_comma,       \
                                 specs_split_comma, \
                                 is_int_literal_constant
from fparser.common.utils import classes, AnalyzeError


class AttributeHolder(object):
    # copied from symbolic.base module
    """
    Defines a object with predefined attributes. Only those attributes
    are allowed that are specified as keyword arguments of a constructor.
    When an argument is callable then the corresponding attribute will
    be read-only and set by the value the callable object returns.
    """
    def __init__(self, **kws):
        self._attributes = {}
        self._readonly = []
        for k, v in list(kws.items()):
            self._attributes[k] = v
            if callable(v):
                self._readonly.append(k)
        return

    def __getattr__(self, name):
        if name not in self._attributes:
            message = '%s instance has no attribute %r, '\
                      + 'expected attributes: %s'
            attributes = ', '.join(list(self._attributes.keys()))
            raise AttributeError(message % (self.__class__.__name__,
                                            name,
                                            attributes))
        value = self._attributes[name]
        if callable(value):
            value = value()
            self._attributes[name] = value
        return value

    def __setattr__(self, name, value):
        if name in ['_attributes', '_readonly']:
            self.__dict__[name] = value
            return
        if name in self._readonly:
            message = '%s instance attribute %r is readonly'
            raise AttributeError(message % (self.__class__.__name__, name))
        if name not in self._attributes:
            message = '%s instance has no attribute %r, ' \
                      + 'expected attributes: %s'
            attributes = ','.join(list(self._attributes.keys()))
            raise AttributeError(message % (self.__class__.__name__,
                                            name,
                                            attributes))
        self._attributes[name] = value

    def isempty(self):
        for k in list(self._attributes.keys()):
            v = getattr(self, k)
            if v:
                return False
        return True

    def __repr__(self):
        return self.torepr()

    def torepr(self, depth=-1, tab=''):
        if depth == 0:
            return tab + self.__class__.__name__
        lines = [self.__class__.__name__+':']
        ttab = tab + '    '
        for k in list(self._attributes.keys()):
            v = getattr(self, k)
            if v:
                if isinstance(v, list):
                    lines.append(ttab + '%s=<%s-list>' % (k, len(v)))
                elif isinstance(v, dict):
                    lines.append(ttab + '%s=<dict with keys %s>'
                                 % (k, list(v.keys())))
                else:
                    lines.append(ttab + '%s=<%s>' % (k, type(v)))
        return '\n'.join(lines)

    def todict(self):
        d = {}
        for k in list(self._attributes.keys()):
            v = getattr(self, k)
            d[k] = v
        return d


def get_base_classes(cls):
    bases = ()
    for c in cls.__bases__:
        bases += get_base_classes(c)
    return bases + cls.__bases__ + (cls, )


class Variable(object, with_metaclass(classes)):
    """
    Variable instance has attributes:
      name
      typedecl
      dimension
      attributes
      intent
      parent - Statement instances defining the variable
    """

    def __init__(self, parent, name):
        self.parent = parent
        self.parents = [parent]
        self.name = name
        self.typedecl = None
        self.dimension = None
        self.bounds = None
        self.length = None
        self.attributes = []
        self.intent = None
        self.bind = []
        self.check = []
        self.init = None

        # after calling analyze the following additional attributes are set:
        # .is_array:
        #    rank
        #    shape
        return

    def __repr__(self):
        line = []
        for a in ['name', 'typedecl', 'dimension', 'bounds', 'length',
                  'attributes', 'intent', 'bind', 'check', 'init']:
            v = getattr(self, a)
            if v:
                line.append('%s=%r' % (a, v))
        return 'Variable: ' + ', '.join(line)

    def get_typedecl(self):
        if self.typedecl is None:
            self.set_type(self.parent.get_type(self.name))
        return self.typedecl

    def add_parent(self, parent):
        if id(parent) not in list(map(id, self.parents)):
            self.parents.append(parent)
        self.parent = parent
        return

    def set_type(self, typedecl):
        if self.typedecl is not None:
            if not self.typedecl == typedecl:
                self.parent.warning(
                    'variable %r already has type %s, '
                    'resetting to %s'
                    % (self.name, self.typedecl.tostr(), typedecl.tostr()))
        assert typedecl is not None
        self.typedecl = typedecl
        return

    def set_init(self, expr):
        if self.init is not None:
            if not self.init == expr:
                self.parent.warning(
                    'variable %r already has initialization %r, '
                    'resetting to %r' % (self.name, self.expr, expr))
        self.init = expr
        return

    def set_dimension(self, dims):
        import six
        dims = [tuple(dim.split(':')) for dim in dims]
        dims = [tuple(map(six.text_type.strip, dim)) for dim in dims]
        if self.dimension is not None:
            if not self.dimension == dims:
                self.parent.warning(
                    'variable %r already has dimension %r, '
                    'resetting to %r' % (self.name, self.dimension, dims))
        self.dimension = dims
        return

    def set_bounds(self, bounds):
        if self.bounds is not None:
            if not self.bounds == bounds:
                self.parent.warning(
                    'variable %r already has bounds %r, '
                    'resetting to %r' % (self.name, self.bounds, bounds))
        self.bounds = bounds
        return

    def set_length(self, length):
        if self.length is not None:
            if not self.length == length:
                self.parent.warning(
                    'variable %r already has length %r, '
                    'resetting to %r' % (self.name, self.length, length))
        self.length = length
        return

    known_intent_specs = ['IN', 'OUT', 'INOUT', 'CACHE', 'HIDE', 'COPY',
                          'OVERWRITE', 'CALLBACK', 'AUX', 'C', 'INPLACE',
                          'OUT=']

    def set_intent(self, intent):
        if self.intent is None:
            self.intent = []
        for i in intent:
            if i not in self.intent:
                if i not in self.known_intent_specs:
                    self.parent.warning('unknown intent-spec %r for %r'
                                        % (i, self.name))
                self.intent.append(i)
        return

    known_attributes = ['PUBLIC', 'PRIVATE', 'ALLOCATABLE', 'ASYNCHRONOUS',
                        'EXTERNAL', 'INTRINSIC', 'OPTIONAL', 'PARAMETER',
                        'POINTER', 'PROTECTED', 'SAVE', 'TARGET', 'VALUE',
                        'VOLATILE', 'REQUIRED']

    def is_intent_in(self):
        # TODO Something hinky is going on here. self.intent is a list which
        #      doesn't make a lot of sense. How can a variable have more
        #      than one intent?
        # TODO Furthermore if no intent is specified the assumed intent is
        #      "inout". Below an explicit "inout" returns False but a None
        #      returns True.
        if not self.intent:
            return True
        if 'HIDE' in self.intent:
            return False
        if 'INPLACE' in self.intent:
            return False
        if 'IN' in self.intent:
            return True
        if 'OUT' in self.intent:
            return False
        if 'INOUT' in self.intent:
            return False
        if 'OUTIN' in self.intent:
            return False
        return True

    def is_intent_inout(self):
        if not self.intent:
            return False
        if 'INOUT' in self.intent:
            if 'IN' in self.intent \
               or 'HIDE' in self.intent \
               or 'INPLACE' in self.intent:
                message = 'INOUT ignored in INPUT(%s)'
                self.warning(message % (', '.join(self.intent)))
                return False
            return True
        return False

    def is_intent_hide(self):
        if not self.intent:
            return False
        if 'HIDE' in self.intent:
            return True
        if 'OUT' in self.intent:
            return 'IN' not in self.intent \
                   and 'INPLACE' not in self.intent \
                   and 'INOUT' not in self.intent
        return False

    def is_intent_inplace(self):
        return self.intent and 'INPLACE' in self.intent

    def is_intent_out(self):
        return self.intent and 'OUT' in self.intent

    def is_intent_c(self):
        return self.intent and 'C' in self.intent

    def is_intent_cache(self):
        return self.intent and 'CACHE' in self.intent

    def is_intent_copy(self):
        return self.intent and 'COPY' in self.intent

    def is_intent_overwrite(self):
        return self.intent and 'OVERWRITE' in self.intent

    def is_intent_callback(self):
        return self.intent and 'CALLBACK' in self.intent

    def is_intent_aux(self):
        return self.intent and 'AUX' in self.intent

    def is_private(self):
        if 'PUBLIC' in self.attributes:
            return False
        if 'PRIVATE' in self.attributes:
            return True
        return self.parent.parent.check_private(self.name)

    def is_public(self):
        return not self.is_private()

    def is_allocatable(self):
        return 'ALLOCATABLE' in self.attributes

    def is_external(self):
        return 'EXTERNAL' in self.attributes

    def is_intrinsic(self):
        return 'INTRINSIC' in self.attributes

    def is_parameter(self):
        return 'PARAMETER' in self.attributes

    def is_optional(self):
        return 'OPTIONAL' in self.attributes \
               and 'REQUIRED' not in self.attributes \
               and not self.is_intent_hide()

    def is_required(self):
        return self.is_optional() and not self.is_intent_hide()

    def is_pointer(self):
        return 'POINTER' in self.attributes

    def is_array(self):
        return not not (self.bounds or self.dimension)

    def is_scalar(self):
        return not self.is_array()

    def update(self, *attrs):
        attributes = self.attributes
        if len(attrs) == 1 and isinstance(attrs[0], (tuple, list)):
            attrs = attrs[0]
        for attr in attrs:
            lattr = attr.lower()
            uattr = attr.upper()
            if lattr.startswith('dimension'):
                assert self.dimension is None, repr((self.dimension, attr))
                line = attr[9:].lstrip()
                assert line[0] + line[-1] == '()', repr(line)
                self.set_dimension(
                        split_comma(line[1:-1].strip(), self.parent.item))
                continue
            if lattr.startswith('intent'):
                line = attr[6:].lstrip()
                assert line[0] + line[-1] == '()', repr(line)
                self.set_intent(specs_split_comma(line[1:-1].strip(),
                                                  self.parent.item,
                                                  upper=True))
                continue
            if lattr.startswith('bind'):
                line = attr[4:].lstrip()
                assert line[0] + line[-1] == '()', repr(line)
                self.bind = specs_split_comma(line[1:-1].strip(),
                                              self.parent.item,
                                              upper=True)
                continue
            if lattr.startswith('check'):
                line = attr[5:].lstrip()
                assert line[0] + line[-1] == '()', repr(line)
                self.check.extend(split_comma(line[1:-1].strip(),
                                              self.parent.item))
                continue
            if uattr not in attributes:
                if uattr not in self.known_attributes:
                    self.parent.warning('unknown attribute %r' % (attr))
                attributes.append(uattr)
        return

    def __str__(self):
        s = ''
        typedecl = self.get_typedecl()
        if typedecl is not None:
            s += typedecl.tostr() + ' '
        a = self.attributes[:]
        if self.dimension is not None:
            dimensions = [':'.join(spec) for spec in self.dimension]
            a.append('DIMENSION(%s)' % (', '.join(dimensions)))
        if self.intent is not None:
            a.append('INTENT(%s)' % (', '.join(self.intent)))
        if self.bind:
            a.append('BIND(%s)' % (', '.join(self.bind)))
        if self.check:
            a.append('CHECK(%s)' % (', '.join(self.check)))
        if a:
            s += ', ' + ', '.join(a) + ' :: '
        s += self.name
        if self.bounds:
            s += '(%s)' % (', '.join([':'.join(spec) for spec in self.bounds]))
        if self.length:
            if is_int_literal_constant(self.length):
                s += '*%s' % (self.length)
            else:
                s += '*(%s)' % (self.length)
        if self.init:
            s += ' = ' + self.init
        return s

    def get_array_spec(self):
        assert self.is_array(), 'array_spec is available only for arrays'
        if self.bounds:
            if self.dimension:
                message = 'both bounds=%r and dimension=%r are defined, ' \
                          + 'ignoring dimension.'
                self.parent.warning(message % (self.bounds, self.dimension))
            array_spec = self.bounds
        else:
            array_spec = self.dimension
        return array_spec

    def is_deferred_shape_array(self):
        if not self.is_array():
            return False
        return self.is_allocatable() or self.is_pointer()

    def is_assumed_size_array(self):
        if not self.is_array():
            return False
        return self.get_array_spec()[-1][-1] == '*'

    def is_assumed_shape_array(self):
        if not self.is_array():
            return False
        if self.is_deferred_shape_array():
            return False
        for spec in self.get_array_spec():
            if not spec[-1]:
                return True
        return False

    def is_explicit_shape_array(self):
        if not self.is_array():
            return False
        if self.is_deferred_shape_array():
            return False
        for spec in self.get_array_spec():
            if not spec[-1] or spec[-1] == '*':
                return False
        return True

    def is_allocatable_array(self):
        return self.is_array() and self.is_allocatable()

    def is_array_pointer(self):
        return self.is_array() and self.is_pointer()

    def analyze(self):
        typedecl = self.get_typedecl()
        if self.is_array():
            array_spec = self.get_array_spec()
            self.rank = len(array_spec)
            if self.is_deferred_shape_array():  # a(:, :)
                pass
            elif self.is_explicit_shape_array():
                shape = []
                for spec in array_spec:
                    if len(spec) == 1:
                        shape.append(spec[0].replace(' ', ''))
                    else:
                        try:
                            # lower subscript
                            lss = int(spec[0].replace(' ', ''))
                            # upper subscript
                            uss = int(spec[1].replace(' ', ''))
                            n = uss - (lss - 1)
                        except ValueError:
                            n = '(%s)-(%s)' % (spec[1], spec[0])
                        shape.append(str(n))
                self.shape = shape
        return

    def error(self, message):
        return self.parent.error(message)

    def warning(self, message):
        return self.parent.warning(message)

    def info(self, message):
        return self.parent.info(message)


class ProgramBlock(object, with_metaclass(classes)):
    pass


class Statement(object, with_metaclass(classes)):
    """
    Statement instance has attributes:
      parent  - Parent BeginStatement or FortranParser instance
      item    - Line instance containing the statement line
      isvalid - boolean, when False, the Statement instance will be ignored
    """

    modes = ['free', 'fix', 'f77', 'pyf']
    _repr_attr_names = []

    def __init__(self, parent, item):
        self.parent = parent
        if item is not None:
            self.reader = item.reader
        else:
            self.reader = parent.reader
        self.top = getattr(parent, 'top', None)  # the top of statement tree
        self.item = item

        if isinstance(parent, ProgramBlock):
            self.programblock = parent
        elif isinstance(self, ProgramBlock):
            self.programblock = self
        elif hasattr(parent, 'programblock'):
            self.programblock = parent.programblock
        else:
            pass

        # When a statement instance is constructed by error, set isvalid to
        # False
        self.isvalid = True
        # when a statement should be ignored, set ignore to True
        self.ignore = False

        # attribute a will hold analyze information.
        a_dict = {}
        for cls in get_base_classes(self.__class__):
            if hasattr(cls, 'a'):
                a_dict.update(copy.deepcopy(cls.a.todict()))
        self.a = AttributeHolder(**a_dict)
        if hasattr(self.__class__, 'a'):
            assert self.a is not self.__class__.a

        self.process_item()

        return

    def __repr__(self):
        return self.torepr()

    def torepr(self, depth=-1, incrtab=''):
        tab = incrtab + self.get_indent_tab()
        clsname = self.__class__.__name__
        lines = [tab + clsname]
        if depth == 0:
            return '\n'.join(lines)
        ttab = tab + '  '
        for n in self._repr_attr_names:
            attr = getattr(self, n, None)
            if not attr:
                continue
            if hasattr(attr, 'torepr'):
                r = attr.torepr(depth-1, incrtab)
            else:
                r = repr(attr)
            lines.append(ttab + '%s=%s' % (n, r))
        if self.item is not None:
            lines.append(ttab + 'item=%r' % (self.item))
        if not self.isvalid:
            lines.append(ttab + 'isvalid=%r' % (self.isvalid))
        if self.ignore:
            lines.append(ttab + 'ignore=%r' % (self.ignore))
        if not self.a.isempty():
            lines.append(ttab
                         + 'a='
                         + self.a.torepr(depth-1, incrtab+'  ').lstrip())
        return '\n'.join(lines)

    def get_indent_tab(self, deindent=False, isfix=None):
        if isfix is None:
            isfix = self.reader.format.is_fixed
        if isfix:
            tab = ' '*6
        else:
            tab = ''
        p = self.parent
        while isinstance(p, Statement):
            tab += '  '
            p = p.parent
        if deindent:
            tab = tab[:-2]
        label = getattr(self.item, 'label', None)
        if label is None:
            return tab
        s = str(label)
        if isfix:
            s = ' '+s
        tab = tab[len(s):]
        if not tab:
            tab = ' '
        tab = s + tab
        return tab

    def __str__(self):
        return self.tofortran()

    def asfix(self):
        lines = []
        for line in self.tofortran(isfix=True).split('\n'):
            if len(line) > 72 and line[0] == ' ':
                lines.append(line[:72]+'&\n     &')
                line = line[72:]
                while len(line) > 66:
                    lines.append(line[:66]+'&\n     &')
                    line = line[66:]
                lines.append(line+'\n')
            else:
                lines.append(line+'\n')
        return ''.join(lines).replace('\n     &\n', '\n')

    def format_message(self, kind, message):
        if self.item is not None:
            message = self.reader.format_message(kind, message,
                                                 self.item.span[0],
                                                 self.item.span[1])
        else:
            return message
        return message

    # def show_message(self, message, stream=sys.stderr):
    #     print >> stream, message
    #     stream.flush()
    #     return

    def error(self, message):
        message = self.format_message('ERROR', message)
        logging.getLogger(__name__).error(message)
        return

    def warning(self, message):
        message = self.format_message('WARNING', message)
        logging.getLogger(__name__).warning(message)
        return

    def info(self, message):
        message = self.format_message('INFO', message)
        logging.getLogger(__name__).info(message)
        return

    def analyze(self):
        self.warning('nothing analyzed')
        return

    def get_variable(self, name):
        """ Return Variable instance of variable name.
        """
        mth = getattr(self, 'get_variable_by_name', self.parent.get_variable)
        return mth(name)

    def get_type(self, name):
        """ Return type declaration using implicit rules
        for name.
        """
        mth = getattr(self, 'get_type_by_name', self.parent.get_type)
        return mth(name)

    def get_type_decl(self, kind):
        mth = getattr(self, 'get_type_decl_by_kind', self.parent.get_type_decl)
        return mth(kind)

    def get_provides(self):
        """
        Returns dictonary containing statements that block provides or None
        when N/A.
        """
        return


class BeginStatement(Statement):
    """[ construct_name : ] <blocktype> [ <name> ]

    BeginStatement instances have additional attributes:
      name
      blocktype

    Block instance has attributes:
      content - list of Line or Statement instances
      name    - name of the block, unnamed blocks are named
                with the line label
      construct_name - name of a construct
      parent  - Block or FortranParser instance
      item    - Line instance containing the block start statement
      get_item, put_item - methods to retrive/submit Line instances
                from/to Fortran reader.
      isvalid - boolean, when False, the Block instance will be ignored.

      stmt_cls, end_stmt_cls

    """
    _repr_attr_names = ['blocktype',
                        'name',
                        'construct_name'] + Statement._repr_attr_names

    def __init__(self, parent, item=None):

        self.content = []
        self.get_item = parent.get_item  # get line function
        self.put_item = parent.put_item  # put line function
        if not hasattr(self, 'blocktype'):
            self.blocktype = self.__class__.__name__.lower()
        if not hasattr(self, 'name'):
            # process_item may change this
            self.name = '__'+self.blocktype.upper()+'__'
        self.construct_name = getattr(item, 'name', None)
        Statement.__init__(self, parent, item)
        return

    def tostr(self):
        return self.blocktype.upper() + ' ' + self.name

    def tofortran(self, isfix=None):
        construct_name = self.construct_name
        construct_name = construct_name + ': ' if construct_name else ''
        lines = [self.get_indent_tab(isfix=isfix)
                 + construct_name + self.tostr()]
        for c in self.content:
            lines.append(c.tofortran(isfix=isfix))
        return '\n'.join(lines)

    def torepr(self, depth=-1, incrtab=''):
        tab = incrtab + self.get_indent_tab()
        ttab = tab + '  '
        lines = [Statement.torepr(self, depth=depth, incrtab=incrtab)]
        if depth == 0 or not self.content:
            return '\n'.join(lines)
        lines.append(ttab+'content:')
        for c in self.content:
            if isinstance(c, EndStatement):
                lines.append(c.torepr(depth-1, incrtab))
            else:
                lines.append(c.torepr(depth-1, incrtab + '  '))
        return '\n'.join(lines)

    def process_item(self):
        """ Process the line
        """
        item = self.item
        if item is None:
            return
        self.fill()
        return

    def fill(self, end_flag=False):
        """
        Fills blocks content until the end of block statement.
        """

        mode = self.reader.format.mode
        class_list = self.get_classes()
        self.classes = [cls for cls in class_list if mode in cls.modes]
        self.pyf_classes = [cls for cls in class_list if 'pyf' in cls.modes]

        item = self.get_item()
        while item is not None:
            if isinstance(item, Line):
                if self.process_subitem(item):
                    end_flag = True
                    break
            elif isinstance(item, Comment):
                # TODO: FIX ME, Comment content is a string
                self.content.append(classes.Comment(self, item))
            else:
                raise NotImplementedError(repr(item))
            item = self.get_item()

        if not end_flag:
            self.warning('failed to find the end of block')
        return

    def process_subitem(self, item):
        """
        Check if item is blocks start statement, if it is, read the block.

        Return True to stop adding items to given block.
        """
        line = item.get_line()

        # First check for the end of block
        cls = self.end_stmt_cls
        if cls.match(line):
            stmt = cls(self, item)
            if stmt.isvalid:
                self.content.append(stmt)
                return True

        if item.is_f2py_directive:
            classes = self.pyf_classes
        else:
            classes = self.classes

        # Look for statement match
        for cls in classes:
            if cls.match(line):
                stmt = cls(self, item)
                if stmt.isvalid:
                    if not stmt.ignore:
                        self.content.append(stmt)
                    return False
                # item may be cloned that changes the items line:
                line = item.get_line()

        # Check if f77 code contains inline comments or other f90
        # constructs that got undetected by get_source_info.
        if item.reader.format.is_f77:
            i = line.find('!')
            if i != -1:
                message = item.reader.format_message(
                        'WARNING',
                        'no parse pattern found for "%s" in %r block, '
                        'trying to remove inline comment (not in Fortran 77).'
                        % (item.get_line(), self.__class__.__name__),
                        item.span[0], item.span[1])
                # .. but at the expense of loosing the comment.
                logging.getLogger(__name__).warning(message)
                if line[:i]:
                    newitem = item.copy(line[:i].rstrip())
                    return self.process_subitem(newitem)
                else:
                    return True

            # try fix statement classes
            f77_classes = self.classes
            classes = []
            for cls in self.get_classes():
                if 'f77' in cls.modes and cls not in f77_classes:
                    classes.append(cls)
            if classes:
                message = item.reader.format_message(
                        'WARNING',
                        'no parse pattern found for "%s" in %r block'
                        ' maybe due to strict f77 mode.'
                        ' Trying f90 fix mode patterns..'
                        % (item.get_line(), self.__class__.__name__),
                        item.span[0], item.span[1])
                logging.getLogger(__name__).warning(message)

                item.reader.set_mode(False, False)
                self.classes = classes

                r = BeginStatement.process_subitem(self, item)
                if r is None:
                    # restore f77 fix mode
                    self.classes = f77_classes
                    item.reader.set_mode(False, True)
                else:
                    message = item.reader.format_message(
                        'INFORMATION',
                        'The f90 fix mode resolved the parse pattern issue.'
                        ' Setting reader to f90 fix mode.',
                        item.span[0], item.span[1])
                    logging.getLogger(__name__).info(message)
                    # set f90 fix mode
                    self.classes = f77_classes + classes
                    self.reader.set_mode(False, False)
                return r

        self.handle_unknown_item(item)
        return

    def handle_unknown_item(self, item):
        '''Called when process_subitem does not find a start or end of block.
        It adds the item (which is an instance of Line) to the content, but
        then raises an AnalyzeError. An instance of Line in content typically
        results in other errors later (e.g. because Line has no analyze
        method).
        '''
        message = item.reader.format_message(
            'WARNING',
            'no parse pattern found for "%s" in %r block.'
            % (item.get_line(), self.__class__.__name__),
            item.span[0], item.span[1])
        logging.getLogger(__name__).warning(message)
        self.content.append(item)
        raise AnalyzeError(message)

    def analyze(self):
        for stmt in self.content:
            stmt.analyze()
        return


class EndStatement(Statement):
    """
    END [<blocktype> [<name>]]

    EndStatement instances have additional attributes:
      name
      blocktype
    """
    _repr_attr_names = ['blocktype', 'name'] + Statement._repr_attr_names

    def __init__(self, parent, item):
        if not hasattr(self, 'blocktype'):
            self.blocktype = self.__class__.__name__.lower()[3:]
        Statement.__init__(self, parent, item)

    def process_item(self):
        item = self.item
        line = item.get_line().replace(' ', '')[3:]
        line = item.apply_map(line)
        blocktype = self.blocktype

        if line.lower().startswith(blocktype):
            line = line[len(blocktype):].strip()
        else:
            if line:
                # not the end of expected block
                line = ''
                self.isvalid = False
        if self.parent.construct_name:
            name = self.parent.construct_name
        else:
            name = self.parent.name
        if line:
            # line variable is already cast to lower case so would fail if any
            # upper case letters exist in the label. Also, fortran is case
            # insensitive anyway so we should assume labels may have a
            # different case and therefore cast both to the same case in our
            # equivalence test.
            if line.lower() != name.lower():
                message = 'expected the end of %r block ' \
                          + 'but got the end of %r, skipping.'
                self.warning(message % (name, line))
                self.isvalid = False
        self.name = name

    def analyze(self):
        return

    def get_indent_tab(self, deindent=False, isfix=None):
        return Statement.get_indent_tab(self, deindent=True, isfix=isfix)

    def tofortran(self, isfix=None):
        '''Returns a valid Fortran string for this END statement. It
        guarantees that there is no white space after the 'END' in case
        of an unnamed statement.

        :param bool isfix: True if the code is in fixed format.

        :returns: the (named or unnamed) valid Fortran END statement \
                  as a string.
        :rtype: str

        '''
        if self.name:
            return self.get_indent_tab(isfix=isfix) + 'END {0} {1}'\
                .format(self.blocktype.upper(), self.name)

        # Make sure there is no space after an unnamed END:
        return self.get_indent_tab(isfix=isfix) + 'END {0}'\
            .format(self.blocktype.upper())



# Copyright (c) 2018 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R430 : This file tests the support for the
Derived Type Statement e.g.

type, private, abstract :: my_type(b,c)

'''

import pytest
from fparser.two.Fortran2003 import Derived_Type_Stmt
from fparser.two.utils import NoMatchError, InternalError


def test_valid(f2003_create):
    '''Check that valid input is parsed correctly.

    '''

    # simple minimal type statement
    ast = Derived_Type_Stmt("type a")
    assert "TYPE :: a" in str(ast)
    assert repr(ast) == "Derived_Type_Stmt(None, Type_Name('a'), None)"

    # simple minimal type statement with spaces
    ast = Derived_Type_Stmt("  type  a  ")
    assert "TYPE :: a" in str(ast)

    # simple minimal type statement with no spaces
    ast = Derived_Type_Stmt("typea")
    assert "TYPE :: a" in str(ast)

    # simple minimal type statement with "::"
    ast = Derived_Type_Stmt("type :: a")
    assert "TYPE :: a" in str(ast)

    # type statement with attribute spec list
    ast = Derived_Type_Stmt("type, private, abstract :: a")
    assert "TYPE, PRIVATE, ABSTRACT :: a" in str(ast)
    assert repr(ast) == ("Derived_Type_Stmt(Type_Attr_Spec_List(',', "
                         "(Access_Spec('PRIVATE'), "
                         "Type_Attr_Spec('ABSTRACT', None))), "
                         "Type_Name('a'), None)")

    # type statement with type parameter name list
    ast = Derived_Type_Stmt("type, private, abstract :: a(b,c)")
    assert "TYPE, PRIVATE, ABSTRACT :: a(b, c)" in str(ast)
    assert repr(ast) == ("Derived_Type_Stmt(Type_Attr_Spec_List(',', "
                         "(Access_Spec('PRIVATE'), "
                         "Type_Attr_Spec('ABSTRACT', None))), "
                         "Type_Name('a'), Type_Param_Name_List(',', "
                         "(Name('b'), Name('c'))))")

    # type statement with type parameter name list spaces
    ast = Derived_Type_Stmt("  type , private , abstract :: a ( b , c )  ")
    assert "TYPE, PRIVATE, ABSTRACT :: a(b, c)" in str(ast)

    # type statement with type parameter name list no spaces
    ast = Derived_Type_Stmt("type,private,abstract::a(b,c)")
    assert "TYPE, PRIVATE, ABSTRACT :: a(b, c)" in str(ast)

    # type statement with type parameter name list no colons
    ast = Derived_Type_Stmt("type a(b,c)")
    assert "TYPE :: a(b, c)" in str(ast)


def test_errors(f2003_create):
    '''Check that invalid input does not match.

    '''

    for value in ["", "  ", "typ", "type", "type ::", "type, ::",
                  "type x ::", "type a b", "type a()", "type a(  )",
                  "type a(b) c"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Derived_Type_Stmt(value)
        assert "Derived_Type_Stmt: '{0}'".format(value) in str(excinfo.value)


def test_tostr_1(f2003_create, monkeypatch):
    '''Check that Derived_Type_Stmt.tostr() raises an exception if there
    is an invalid number of items.

    '''

    ast = Derived_Type_Stmt("type a")
    monkeypatch.setattr(ast, "items", ["A"])
    with pytest.raises(InternalError) as excinfo:
        str(ast)
    assert "should be of size 3 but found '1'" in str(excinfo.value)


def test_tostr_2(f2003_create, monkeypatch):
    '''Check that Derived_Type_Stmt.tostr() raises an exception if the
    content of items[1] is invalid.

    '''

    ast = Derived_Type_Stmt("type a")
    monkeypatch.setattr(ast, "items", [None, None, None])
    with pytest.raises(InternalError) as excinfo:
        str(ast)
    assert ("'items[1]' should be a Name instance containing the "
            "derived type name but it is empty") in str(excinfo.value)


def test_get_start_name(f2003_create):
    '''Check that the appropriate name is returned from
    get_start_name()

    '''
    ast = Derived_Type_Stmt("type a")
    assert ast.get_start_name() == "a"



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Tests for a Fortran 2003 data edit descriptor that only allows
descriptors that conform to (the first clause in) constraint C1002.

'''

import pytest
from fparser.two.Fortran2003 import Data_Edit_Desc_C1002
from fparser.two.utils import NoMatchError, InternalError


def test_wd(f2003_create):
    '''Check that valid w.d format specifications are parsed
    correctly. Also include an example with spaces and multiple
    digits with spaces.

    '''
    for descriptor in ["F", "E", "EN", "ES", "G", "D", "f", "e", "en", "En",
                       "eN", "es", "g", "d"]:
        for my_input in ["{0}2.3".format(descriptor),
                         " {0} 2 . 3 ".format(descriptor),
                         " {0} 2 2 . 3 3 ".format(descriptor)]:
            ast = Data_Edit_Desc_C1002(my_input)
            assert str(ast) == my_input.upper().replace(" ", "")


def test_wde(f2003_create):
    '''Check that valid w.dEe format specifications are parsed
    correctly. Also include an example with spaces and multiple
    digits with spaces.

    '''
    for descriptor in ["E", "EN", "ES", "G", "e", "en", "En", "eN", "es", "g"]:
        for my_input in ["{0}2.3E4".format(descriptor),
                         " {0} 2 . 3 E 4 ".format(descriptor),
                         " {0} 2 2 . 3 3 E 4 4 ".format(descriptor)]:
            ast = Data_Edit_Desc_C1002(my_input)
            assert str(ast) == my_input.upper().replace(" ", "")


def test_errors(f2003_create):
    '''Check that format specifications not conforming to the C1002 1st
    clause, raise an NoMatchError exception.

    '''
    for my_input in [None, "", "  ", "I2", "F2.3E4", "D2.3E4", "F", "D",
                     "ES", "F2", "F2.", "E2E3", "E2.E3", "E2.3E", "E2.3E4E"]:
        with pytest.raises(NoMatchError):
            _ = Data_Edit_Desc_C1002(my_input)


def test_internal_error1(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the length of the Items
    list is not 4 as the str() method assumes that it is.

    '''
    my_input = "E1.2E3"
    ast = Data_Edit_Desc_C1002(my_input)
    monkeypatch.setattr(ast, "items", [None, None, None])
    with pytest.raises(InternalError) as excinfo:
        str(ast)
    assert "has '3' items, but expecting 4." in str(excinfo.value)


def test_internal_error2(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the descriptor name
    (entry 0 of items) is empty or None as the str() method assumes
    that it is a string with content.

    '''
    my_input = "E1.2E3"
    ast = Data_Edit_Desc_C1002(my_input)
    for content in [None, ""]:
        monkeypatch.setattr(ast, "items", [content]+list(ast.items[1:]))
        with pytest.raises(InternalError) as excinfo:
            str(ast)
        assert "should be a descriptor name but is empty or None" \
            in str(excinfo.value)


def test_internal_error3(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the w value
    (entry 1 of items) is empty or None as the str() method assumes
    that it is a string with content.

    '''
    my_input = "E1.2E3"
    ast = Data_Edit_Desc_C1002(my_input)
    for content in [None, ""]:
        monkeypatch.setattr(ast, "items", [ast.items[0], content,
                                           ast.items[2], ast.items[3]])
        with pytest.raises(InternalError) as excinfo:
            str(ast)
        assert "should be the w value but is empty or None" in \
            str(excinfo.value)


def test_internal_error4(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the m value (entry 2 of
    items) is empty or None as the str() method assumes that it is a
    string with content.

    '''
    my_input = "E1.2E3"
    ast = Data_Edit_Desc_C1002(my_input)
    for content in [None, ""]:
        monkeypatch.setattr(ast, "items", [ast.items[0], ast.items[1],
                                           content, ast.items[3]])
        with pytest.raises(InternalError) as excinfo:
            str(ast)
        assert "should be the m value but is empty or None" in \
            str(excinfo.value)


def test_internal_error5(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the e value (entry 3 of
    items) has content when the descriptor is F or D as it should be
    None.

    '''
    for my_input in ["F1.2", "D1.2"]:
        ast = Data_Edit_Desc_C1002(my_input)
        monkeypatch.setattr(ast, "items", list(ast.items[0:3])+["3"])
        with pytest.raises(InternalError) as excinfo:
            str(ast)
        assert ("has an exponent value '3' but this is not allowed for 'F' "
                "and 'D' descriptors" in str(excinfo.value))


def test_internal_error6(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the descriptor name
    has an unexpected value.

    '''
    my_input = "E1.2E3"
    ast = Data_Edit_Desc_C1002(my_input)
    monkeypatch.setattr(ast, "items", ["INVALID"]+list(ast.items[1:]))
    with pytest.raises(InternalError) as excinfo:
        str(ast)
    assert "Unexpected descriptor name 'INVALID'" in str(excinfo.value)



# Copyright (c) 2020 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' Test the walk() routine provided by utils.py. '''

import pytest
from fparser.api import get_reader
from fparser.two.utils import walk
from fparser.two import Fortran2003


@pytest.mark.usefixtures("f2003_create")
def test_walk():
    ''' Test the walk() utility. '''
    reader = get_reader("program hello\n"
                        "write(*,*) 'hello'\n"
                        "write(*,*) 'goodbye'\n"
                        "end program hello\n")
    main = Fortran2003.Program(reader)
    # Check that walk produces the same result whether or not the
    # starting node is in a list.
    all_nodes = walk(main)
    all_nodes2 = walk([main])
    assert all_nodes == all_nodes2
    # Check that we can pull out nodes of a particular type
    all_writes = walk(main, types=Fortran2003.Write_Stmt)
    assert len(all_writes) == 2
    assert "hello" in str(all_writes[0])
    assert "goodbye" in str(all_writes[1])
    # Walk from a list of sibling nodes
    io_units = walk(all_writes, types=Fortran2003.Io_Unit)
    assert len(io_units) == 2
    # Should get empty list if no matching nodes found
    node_list = walk(all_writes, Fortran2003.Execution_Part)
    assert node_list == []


@pytest.mark.usefixtures("f2003_create")
def test_walk_debug(capsys):
    ''' Test the debug output of the walk() utility. '''
    import six
    reader = get_reader("program just_a_test\n"
                        "if(.true.)then\n"
                        "  b = 1\n"
                        "end if\n"
                        "end program just_a_test\n")
    main = Fortran2003.Program(reader)
    _ = walk(main, debug=True)
    stdout, _ = capsys.readouterr()
    if six.PY2:
        # Output of capsys under Python 2 is not the same as under 3
        assert stdout.startswith("('child type = ")
    else:
        assert stdout.startswith("child type = ")
    assert "Main_Program" in stdout
    assert "If_Construct" in stdout
    assert "Assignment" in stdout
    assert "Int_Literal_Constant" in stdout
    assert "End_If_Stmt" in stdout
    # Walk only part of the tree and specify an indent
    if_constructs = walk(main, Fortran2003.If_Construct)
    _ = walk(if_constructs[0], indent=4, debug=True)
    stdout, _ = capsys.readouterr()
    if six.PY2:
        # Output of capsys under Python 2 is not the same as under 3
        assert stdout.startswith("('" + 8*" " + "child type =")
    else:
        assert stdout.startswith(8*" "+"child type =")
    assert "Program" not in stdout



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' File containing unit tests for the BlockBase baseclass in utils.py '''

import pytest

from fparser.api import get_reader
from fparser.two.utils import BlockBase
import fparser.two.Fortran2003 as F2003

# TODO #179: full testing of this class. We currently only test the
# comment and include support.


def test_include(f2003_create):
    '''Test the BlockBase match method supports include statements and
    comments before the class that indicates the start of the block
    (the start class).

    '''
    # Use the Main Program rule R1101 as an example
    startcls = F2003.Program_Stmt
    subclasses = [F2003.Specification_Part, F2003.Execution_Part,
                  F2003.Internal_Subprogram_Part]
    endcls = F2003.End_Program_Stmt
    reader = get_reader((
        "include '1'\n"
        "! comment1\n"
        "program test\n"
        "include '2'\n"
        "! comment2\n"
        "integer :: i\n"
        "include '3'\n"
        "! comment3\n"
        "i=1\n"
        "include '4'\n"
        "! comment4\n"
        "contains\n"
        "include '5'\n"
        "! comment5\n"
        "end program test\n"
        "! I should be ignored"
        "include 'so should I'"), ignore_comments=False)
    result = BlockBase.match(startcls, subclasses, endcls, reader)
    assert (
        "([Include_Stmt(Include_Filename('1')), Comment('! comment1'), "
        "Program_Stmt('PROGRAM', Name('test')), Specification_Part("
        "Implicit_Part(Include_Stmt(Include_Filename('2')), "
        "Comment('! comment2')), Type_Declaration_Stmt(Intrinsic_Type_Spec("
        "'INTEGER', None), None, Entity_Decl_List(',', (Entity_Decl(Name("
        "'i'), None, None, None),))), Implicit_Part(Include_Stmt("
        "Include_Filename('3')), Comment('! comment3'))), Execution_Part("
        "Assignment_Stmt(Name('i'), '=', Int_Literal_Constant('1', None)), "
        "Include_Stmt(Include_Filename('4')), Comment('! comment4')), "
        "Internal_Subprogram_Part(Contains_Stmt('CONTAINS'), Include_Stmt("
        "Include_Filename('5')), Comment('! comment5')), End_Program_Stmt("
        "'PROGRAM', Name('test'))],)" in str(result).replace("u'", "'"))
    assert "should" not in str(result)


@pytest.mark.parametrize('strict_order', [True, False])
def test_strict_order_invalid_code(f2003_create, strict_order):
    '''Check that the strict_order flag toggles the parse behaviour as
    expected.
    '''
    subclasses = [F2003.Specification_Part, F2003.Execution_Part]
    reader = get_reader("""
        program main
            i = 2
            integer :: i
        end program main
        """)

    expected = remove_indentation("""([
      Program_Stmt('PROGRAM', Name('main')),
      Execution_Part(
          Assignment_Stmt(Name('i'), '=',
          Int_Literal_Constant('2', None))),
      Specification_Part(
          Type_Declaration_Stmt(
              Intrinsic_Type_Spec('INTEGER', None),
              None,
              Entity_Decl_List(
                  ',',
                  (Entity_Decl(Name('i'), None, None, None),)))),
      End_Program_Stmt('PROGRAM', Name('main'))
     ],)
    """)

    result = BlockBase.match(
        F2003.Program_Stmt, subclasses, F2003.End_Program_Stmt, reader,
        strict_order=strict_order)

    if strict_order:
        assert result is None
    else:
        assert str(result) == expected


def test_strict_order_valid_code(f2003_create):
    '''Tests that the strict_order keyword allows repeated types.
    '''
    subclasses = [F2003.Specification_Part, F2003.Execution_Part]
    reader = get_reader("""
        program main
            integer :: i
            real    :: rho
            i = 2
            rho = i * 3.14
        end program main
        """)

    expected = remove_indentation("""([
      Program_Stmt('PROGRAM', Name('main')),
      Specification_Part(
          Type_Declaration_Stmt(
              Intrinsic_Type_Spec('INTEGER', None), None,
              Entity_Decl_List(
                  ',',
                  (Entity_Decl(Name('i'), None, None, None),))),
          Type_Declaration_Stmt(
              Intrinsic_Type_Spec('REAL', None), None,
              Entity_Decl_List(
                  ',',
                  (Entity_Decl(Name('rho'), None, None, None),)))),
      Execution_Part(
          Assignment_Stmt(Name('i'), '=', Int_Literal_Constant('2', None)),
          Assignment_Stmt(
              Name('rho'), '=', Add_Operand(Name('i'), '*',
              Real_Literal_Constant('3.14', None)))),
      End_Program_Stmt('PROGRAM', Name('main'))
     ],)
    """)
    result = BlockBase.match(
        F2003.Program_Stmt, subclasses, F2003.End_Program_Stmt, reader,
        strict_order=True)

    assert str(result) == expected


def remove_indentation(string):
    '''
    A utility function that removes indented multiline strings of reprs.

    :param str string: the string to dedent.
    :returns: a dedented version of the string.
    :rtype: str

    '''
    # Note, unlike textwrap.dedent this function removes the leading
    # whitespace on each line in a context sensitive way.
    block_indent = 0
    result = ''
    for line in string.split('\n'):
        this_indent = len(line) - len(line.lstrip())
        line = line.lstrip()
        if this_indent > block_indent or not result.endswith(','):
            result += line
        else:
            result += ' ' + line
        block_indent = this_indent
    return result.strip()



# Copyright (c) 2020 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' Test the get_child routine provided by utils.py. '''

import pytest
from fparser.api import get_reader


@pytest.mark.usefixtures("f2003_create")
def test_get_child():
    ''' Test the get_child() utility. '''
    from fparser.two import Fortran2003
    from fparser.two.utils import walk, get_child
    reader = get_reader("program hello\n"
                        "write(*,*) 'hello'\n"
                        "write(*,*) 'goodbye'\n"
                        "end program hello\n")
    main = Fortran2003.Program(reader)
    prog = get_child(main, Fortran2003.Main_Program)
    exe = get_child(prog, Fortran2003.Execution_Part)
    assert isinstance(exe, Fortran2003.Execution_Part)
    write_stmt = get_child(exe, Fortran2003.Write_Stmt)
    # Check that we got the first write and not the second
    assert "goodbye" not in str(write_stmt)
    # The top level has no Io_Control_Spec children
    assert not get_child(main, Fortran2003.Io_Control_Spec)
    # Check functionality when node has children in `items` and
    # not in `content`
    io_nodes = walk(main.content, Fortran2003.Io_Control_Spec)
    assert not hasattr(io_nodes[0], "content")
    io_unit = get_child(io_nodes[0], Fortran2003.Io_Unit)
    assert isinstance(io_unit, Fortran2003.Io_Unit)
    missing = get_child(io_nodes[0], Fortran2003.Execution_Part)
    assert missing is None



# Copyright (c) 2018 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R754 : This file tests the support for the
forall-header statement.

'''

import pytest
from fparser.two.Fortran2003 import Forall_Header
from fparser.two.utils import NoMatchError, InternalError


def test_triplet_list(f2003_create):
    '''Test that one or more triplets is parsed correctly.'''
    for my_input in ["(i=1:n)",
                     "  (  i  =  1  :  n  )  ",
                     "(I=1:N)",
                     "(i=1:n,j=1:n)",
                     "(i=1:n:m,j=1:n)",
                     "(i=1:n,j=1:n:m)",
                     "(i=1:n:m,j=1:n:m)",
                     "(i=1:n:m,j=1:n:m,k=1:n:m)",
                     "(i=a(p,p):n:m,j=1:b(p,p):m,k=1:n:c(p,p))"]:
        ast = Forall_Header(my_input)
        assert my_input.replace(" ", "") in str(ast).replace(" ", "")


def test_mask_expr(f2003_create):
    '''Test that an optional mask expression after a triplet is parsed
    correctly.

    '''
    for my_input in ["(i=1:n,i<4)",
                     "  (  i  =  1  :  n  ,  i  <  4  )  ",
                     "(I=1:N,I<4)",
                     "(i=1:n,j=1:n,i/=j)",
                     "(i=1:n:m,j=1:n,i/=j)",
                     "(i=1:n,j=1:n:m,i/=j)",
                     "(i=1:n:m,j=1:n:m,i/=j)",
                     "(i=1:n:m,j=1:n:m,k=1:n:m,i/=k)",
                     "(i=1:n:m,j=1:n:m,k=1:n:m,a(i,j,k)==0.0)"]:
        ast = Forall_Header(my_input)
        assert my_input.replace(" ", "") in str(ast).replace(" ", "")


def test_parse_error(f2003_create):
    '''Test that NoMatchError is raised for various syntax errors.'''
    for my_input in ["",
                     "  "
                     "(",
                     "()",
                     "i=1:n)",
                     "(i=1:n",
                     "(error)"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Forall_Header(my_input)
        assert "Forall_Header: '{0}'".format(my_input) in str(excinfo.value)


def test_internal_error1(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the length of the Items
    list is not 2 as the str() method assumes that it is.

    '''
    my_input = "(i=1:2)"
    ast = Forall_Header(my_input)
    monkeypatch.setattr(ast, "items", [None, None, None])
    with pytest.raises(InternalError) as excinfo:
        str(ast)
    assert "should be of size 2 but found '3'" in str(excinfo.value)


def test_use_internal_error2(f2003_create, monkeypatch):
    '''Check that an internal error is raised if the triplet lisy (entry 0
    of Items) is empty or None as the str() method assumes that it
    returns a string with content.

    '''
    my_input = "(i=1:2)"
    ast = Forall_Header(my_input)
    for content in [None, ""]:
        monkeypatch.setattr(ast, "items", [content, None])
        with pytest.raises(InternalError) as excinfo:
            str(ast)
        assert ("'items[0]' should be a Forall_Triplet_Spec_List instance "
                "but it is empty") in str(excinfo.value)



# Copyright (c) 2018-2020 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

''' File containing unit tests for the WORDClsBase baseclass in utils.py '''

import pytest
from fparser.two.utils import WORDClsBase, NoMatchError
from fparser.two.Fortran2003 import Name

# test WORDClsBase

# TODO: match: multiple matches (list of values)
# TODO: match: code when the input is not a string or list of values
# TODO: test tostr and tostr_a


def test_wordclsbase():
    '''Test the wordclsbase match method with no optional arguments.'''

    token = "TOKEN"
    name = "name"

    # token only, empty text no match
    text = ""
    result = WORDClsBase.match(token, Name, text)
    assert not result

    # token only, name too short no match
    text = "TOKE"
    result = WORDClsBase.match(token, Name, text)
    assert not result

    # token only, wrong name no match
    text = "wrong"
    result = WORDClsBase.match(token, Name, text)
    assert not result

    # token only, match
    text = token
    result = WORDClsBase.match(token, Name, text)
    assert str(result) == "('{0}', None)".format(token)

    # token only, mixed case match
    text = "ToKeN"
    result = WORDClsBase.match(token, Name, text)
    assert str(result) == "('{0}', None)".format(token)

    # token only, spaces match
    text = "  {0}  ".format(token)
    result = WORDClsBase.match(token, Name, text)
    assert str(result) == "('{0}', None)".format(token)

    # token and name no space no match
    text = "{0}{1}".format(token, name)
    result = WORDClsBase.match(token, Name, text)
    assert not result

    # token and name match
    text = "{0} {1}".format(token, name)
    result = WORDClsBase.match(token, Name, text)
    assert str(result) == "('{0}', Name('{1}'))".format(token, name)

    # token, :: and name no spaces no match
    text = "{0}::{1}".format(token, name)
    with pytest.raises(NoMatchError) as excinfo:
        result = WORDClsBase.match(token, Name, text)
    assert "Name: '::name'" in str(excinfo.value)

    # token, :: and name with spaces no match
    text = "{0} :: {1}".format(token, name)
    with pytest.raises(NoMatchError) as excinfo:
        result = WORDClsBase.match(token, Name, text)
    assert "Name: ':: name'" in str(excinfo.value)


@pytest.mark.xfail(reason="multiple names are treated as one and concatenated")
def test_wordclsbase_multi_names():
    '''Create a separate test as these examples fail as the Name class
    happily strips out spaces. I need to work out whether this is an
    issue for this class or for the Name class itself.

    Test with 'require_cls' is 'False' (the default) which means that
    content for the class ('Name' in this case) is optional and test
    with 'require_cls' is 'True' which means that valid content should
    exist for the class.

    '''

    token = "TOKEN"
    name = "name"

    # token and names no match
    text = "{0} {1} {1}".format(token, name)
    result = WORDClsBase.match(token, Name, text)
    assert not result

    # token and names no match
    text = "{0} {1} {1}".format(token, name)
    result = WORDClsBase.match(token, Name, text, require_cls=True)
    assert not result


def test_wordclsbase_optional_colons():
    '''Test the wordclsbase match method with colons as an optional
    argument. No need to check in combination with the require_cls
    optional argument as the code is independent.

    '''
    token = "TOKEN"
    name = "name"

    # token only - match
    text = token
    result = WORDClsBase.match(token, Name, text, colons=True)
    assert str(result) == "('{0}', None)".format(token)

    # token then name - match
    text = "{0} {1}".format(token, name)
    result = WORDClsBase.match(token, Name, text, colons=True)
    assert str(result) == "('{0}', Name('{1}'))".format(token, name)

    # token then :: (no spaces) with no following name - no match
    text = "{0}::".format(token)
    result = WORDClsBase.match(token, Name, text, colons=True)
    assert not result

    # token then :: with no following name - no match
    text = "{0} :: ".format(token)
    result = WORDClsBase.match(token, Name, text, colons=True)
    assert not result

    # token, :: then name (no spaces) - match
    text = "{0}::{1}".format(token, name)
    result = WORDClsBase.match(token, Name, text, colons=True)
    assert str(result) == "('{0}', Name('{1}'))".format(token, name)

    # token, :: then name - match
    text = "{0} :: {1}".format(token, name)
    result = WORDClsBase.match(token, Name, text, colons=True)
    assert str(result) == "('{0}', Name('{1}'))".format(token, name)


def test_wordclsbase_require_cls():
    '''Test the wordclsbase match method with require_cls as an optional
    argument.

    '''
    token = "TOKEN"
    name = "name"

    # token only - no match
    text = token
    result = WORDClsBase.match(token, Name, text, require_cls=True)
    assert not result

    # token then :: (no spaces) with no following name - no match
    text = "{0}::".format(token)
    result = WORDClsBase.match(token, Name, text, colons=True,
                               require_cls=True)
    assert not result

    # token then :: with no following name - no match
    text = "{0} :: ".format(token)
    result = WORDClsBase.match(token, Name, text, colons=True,
                               require_cls=True)
    assert not result

    # token then name - match
    text = "{0} {1}".format(token, name)
    result = WORDClsBase.match(token, Name, text, colons=True,
                               require_cls=True)
    assert str(result) == "('{0}', Name('{1}'))".format(token, name)

    # token, :: then name (no spaces) - match
    text = "{0}::{1}".format(token, name)
    result = WORDClsBase.match(token, Name, text, colons=True,
                               require_cls=True)
    assert str(result) == "('{0}', Name('{1}'))".format(token, name)

    # token, :: then name - match
    text = "{0} :: {1}".format(token, name)
    result = WORDClsBase.match(token, Name, text, colons=True,
                               require_cls=True)
    assert str(result) == "('{0}', Name('{1}'))".format(token, name)



#!/usr/bin/env python

# Modified work Copyright (c) 2017-2020 Science and Technology
# Facilities Council.
# Original work Copyright (c) 1999-2008 Pearu Peterson.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""Setup script. Used by easy_install and pip."""

from setuptools import setup, find_packages

PACKAGES = find_packages(where="src")

NAME = 'fparser'
AUTHOR = 'Andrew Porter'
AUTHOR_EMAIL = 'trackstand.andy@gmail.com'
URL = 'https://github.com/stfc/fparser'
DOWNLOAD_URL = 'https://github.com/stfc/fparser'
DESCRIPTION = 'The fparser Project'
LONG_DESCRIPTION = '''\
The fparser project is created to develop a parser for
Fortran 77..2008 code. It is based on the work of Pearu Peterson in
the F2PY project (http://www.f2py.com).

See https://github.com/stfc/fparser for more information.
'''
LICENSE = 'OSI Approved :: BSD 3-Clause License'

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'Natural Language :: English',
    'Programming Language :: Fortran',
    'Programming Language :: Python',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development',
    'Topic :: Utilities',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Operating System :: MacOS']

MAJOR = 0
MINOR = 0
MICRO = 11
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)

if __name__ == '__main__':

    setup(
        name=NAME,
        version=VERSION,
        author=AUTHOR,
        author_email=(AUTHOR_EMAIL),
        license=LICENSE,
        url=URL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        classifiers=CLASSIFIERS,
        packages=PACKAGES,
        package_dir={"": "src"},
        install_requires=['six'],
        entry_points={
            'console_scripts': [
                'fparser2=fparser.scripts.fparser2:main',
            ],
        },
        # We need the following line to ensure we get the fparser/log.config
        # file installed.
        include_package_data=True)



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran 2003 rule R311 : This file tests the support for a
defined operator.

'''

import pytest
from fparser.two.Fortran2003 import Defined_Operator
from fparser.two.utils import NoMatchError


def test_defined_operator(f2003_create):
    '''Check that correct defined operator input is parsed correctly. No
    need to test all options as they will be tested by the
    subclasses.

    '''
    # defined unary or binary op
    myinput = ".inv."
    ast = Defined_Operator(myinput)
    assert myinput.upper() in str(ast)
    assert repr(ast) == "Defined_Op('{0}')".format(myinput.upper())
    # extended intrinsic op
    myinput = "**"
    ast = Defined_Operator(myinput)
    assert myinput.upper() in str(ast)
    assert repr(ast) == "Extended_Intrinsic_Op('{0}')".format(myinput)


def test_parse_errors(f2003_create):
    '''The full set of errors will be checked by the subclass tests.  We
    just check a few here to make sure the correct NoMatchError
    exception is raised at this level.

    '''
    for myinput in ["", "  ", ".inv", "inv.", ".false.", "***"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Defined_Operator(myinput)
        assert "Defined_Operator: '{0}'".format(myinput) in str(excinfo.value)



# Modified work Copyright (c) 2017 Science and Technology Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

from fparser import api

def test_reproduce_issue():
    source_str = '''\
      subroutine bndfp()
      use m_struc_def
C-
C 
C
C
C
C
      end
'''
    tree = api.get_reader(source_str, isfree=False, isstrict=False,
                          ignore_comments=False)
    tree = list(tree)
    s, u, c, e = tree[:3]+tree[-1:]
    assert s.span==(1,1),repr(s.span)
    assert u.span==(2,2),repr(u.span)
    assert c.span==(3,3),repr(c.span)
    assert e.span==(9,9),repr(e.span)


def test_reproduce_issue_fix77():
    source_str = '''\
      subroutine foo()
      real a
c
c
      end
'''
    tree = api.get_reader(source_str, isfree=False, isstrict=True,
                          ignore_comments=False)
    tree = list(tree)
    foo, a, comment, end = tree[:3]+tree[-1:]
    assert foo.span == (1, 1)
    assert a.span == (2, 4), repr(a)
    assert comment.span == (5, 5), repr(comment.span)
    assert end.span == (5, 5), repr(end.span)


def test_reproduce_issue_fix90():
    source_str = '''\
      subroutine foo()
      real a
c 1
c 2
      end
'''
    tree = api.get_reader(source_str, isfree=False, isstrict=False)
    tree = list(tree)
    foo, a, comment,end = tree[:3]+tree[-1:]
    assert foo.span==(1,1)
    assert a.span==(2,2),repr(a.span)
    assert end.span==(5,5),repr(end.span)

    source_str = '''\
      subroutine foo()
      real a
c-
c
      end
'''
    tree = api.get_reader(source_str, isfree=False, isstrict=False,
                          ignore_comments=False)
    tree = list(tree)
    foo, a, comment,end = tree[:3]+tree[-1:]
    assert foo.span==(1,1)
    assert a.span==(2,2),repr(a.span)
    assert end.span==(5,5),repr(end.span)

    source_str = '''\
      subroutine foo()
      real a
c
c
      end
'''
    tree = api.get_reader(source_str, isfree=False, isstrict=False,
                          ignore_comments=False)
    tree = list(tree)
    foo, a, comment, end = tree[:3]+tree[-1:]
    assert foo.span==(1,1)
    assert a.span==(2,2),repr(a.span)
    assert comment.span == (3,3)
    assert end.span==(5,5),repr(end.span)

def test_comment_cont_fix90():
    source_str = '''\
          subroutine foo()
      real
c 1
     & a
c 2
      end
'''
    tree = api.get_reader(source_str, isfree=False, isstrict=False,
                          ignore_comments=False)
    tree = list(tree)
    foo, a, comment, end = tree[:3]+tree[-1:]
    assert foo.span==(1,1)
    assert a.span==(2,4),repr(a.span)
    assert comment.span==(3,3),repr(comment.span)
    assert end.span==(6,6)

    source_str = '''\
          subroutine foo()
      real
c
     & a
c 2
      end
'''
    tree = api.get_reader(source_str, isfree=False, isstrict=False,
                          ignore_comments=False)
    tree = list(tree)
    foo, a, comment, end = tree[:3]+tree[-1:]
    assert foo.span==(1,1)
    assert a.span==(2,4),repr(a.span)
    assert comment.span==(3,3),repr(comment.span)
    assert end.span==(6,6)

    source_str = '''\
          subroutine foo()
      real
c 1
     & a
c
      end
'''
    tree = api.get_reader(source_str, isfree=False, isstrict=False,
                          ignore_comments=False)
    tree = list(tree)
    foo, a, comment, end = tree[:3]+tree[-1:]
    assert foo.span==(1,1)
    assert a.span==(2,4),repr(a.span)
    assert comment.span==(3,3),repr(comment.span)
    assert end.span==(6,6)

    source_str = '''\
          subroutine foo()
      real
c 1
     & a
c 2
     &,b
      end
'''
    tree = api.get_reader(source_str, isfree=False, isstrict=False,
                          ignore_comments=False)
    tree = list(tree)
    foo, ab, comment, end = tree[:3]+tree[-1:]
    assert foo.span==(1,1)
    assert ab.span==(2,6),repr(a.span)
    assert comment.span==(3,3),repr(comment.span)
    assert end.span==(7,7)



# Modified work Copyright (c) 2017-2018 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Test fparser support for parsing select-case and select-type blocks

"""

import pytest
import fparser.common.sourceinfo


# We need to monkeypatch the logger used by fparser because it grabs
# stdout before the pytest framework can intercept it. In python < 3
# you can't make a lambda out of 'print' because print is not a
# function (you cannot do "x = print y" for instance). Therefore we
# have to create our own function that simply wraps print and returns
# a value.
def print_wrapper(arg):
    ''' A wrapper that allows us to call print as a function. Used for
    monkeypatching logging calls. '''
    print(arg)
    return None


def test_case():
    ''' Basic tests for parsing of individual case statements '''
    from fparser.one.tests.test_parser import parse
    from fparser.one.block_statements import Case
    assert parse(Case, 'case (1)') == 'CASE ( 1 )'
    assert parse(Case, 'case (1:)') == 'CASE ( 1 : )'
    assert parse(Case, 'case (:1)') == 'CASE ( : 1 )'
    assert parse(Case, 'case (1:2)') == 'CASE ( 1 : 2 )'
    assert parse(Case, 'case (a(1,2))') == 'CASE ( a(1,2) )'
    assert parse(Case, 'case ("ab")') == 'CASE ( "ab" )'
    assert parse(Case, 'case default') == 'CASE DEFAULT'
    assert parse(Case, 'case (1:2 ,3:4)') == 'CASE ( 1 : 2, 3 : 4 )'
    assert parse(Case, 'case (a(1,:):)') == 'CASE ( a(1,:) : )'
    assert parse(Case, 'case default') == 'CASE DEFAULT'


def test_case_internal_error(monkeypatch, capsys):
    ''' Check that expected errors are raised when invalid case
    statements are encountered '''
    from fparser.one.block_statements import Case
    from fparser.common.readfortran import FortranStringReader
    reader = FortranStringReader('CASE (yes)')
    reader.set_format(fparser.common.sourceinfo.FortranFormat(True, False))
    item = next(reader)
    stmt = Case(item, item)
    # Monkeypatch our valid Case object so that get_line() now
    # returns something invalid. We have to do it this way
    # because if we started with this text then we wouldn't get
    # past the match() method
    monkeypatch.setattr(stmt.item, "get_line",
                        lambda: "case invalid")
    # Monkeypatch the Case object so that a call to self.warning
    # (which normally results in a call to the logger) gets replaced
    # with a call to our print_wrapper() function
    monkeypatch.setattr(stmt, "warning", print_wrapper)
    stmt.process_item()
    output, _ = capsys.readouterr()
    print(output)
    assert "Internal error when parsing CASE statement" in output


def test_class_internal_error(monkeypatch, capsys):
    ''' Check that expected errors are raised when invalid CLASS
    statements are encountered '''
    from fparser.one.block_statements import ClassIs
    from fparser.common.readfortran import FortranStringReader
    reader = FortranStringReader('CLASS IS (yes)')
    reader.set_format(fparser.common.sourceinfo.FortranFormat(True, False))
    item = next(reader)
    stmt = ClassIs(item, item)
    # Monkeypatch our valid Case object so that get_line() now
    # returns something invalid. We have to do it this way
    # because if we started with this text then we wouldn't get
    # past the match() method
    monkeypatch.setattr(stmt.item, "get_line",
                        lambda: "class invalid")
    # Monkeypatch the Case object so that a call to self.warning
    # (which normally results in a call to the logger) gets replaced
    # with a call to our print_wrapper() function
    monkeypatch.setattr(stmt, "warning", print_wrapper)
    stmt.process_item()
    output, _ = capsys.readouterr()
    assert "Internal error when parsing CLASS statement" in output


def test_select_case():
    '''Test that fparser correctly recognises select case'''
    from fparser import api
    source_str = '''
    subroutine foo
    integer :: iflag = 1
    real    :: aval = 0.0
    select case(iflag)
    case(1)
      aval = 1.0
    case default
      aval = 0.0
    end select
    ! Same again but with more white space
    select   case(iflag)
    case ( 1 )
      aval = 1.0
    case  default
      aval = 0.0
    end select
    end subroutine foo
    '''
    tree = api.parse(source_str, isfree=True, isstrict=False,
                     ignore_comments=True)
    assert tree
    select_list = []
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectCase):
            select_list.append(statement)
    assert len(select_list) == 2
    for statement in select_list:
        assert isinstance(statement, fparser.one.block_statements.SelectCase)
        assert isinstance(statement.content[0], fparser.one.statements.Case)
        assert isinstance(statement.content[2], fparser.one.statements.Case)
        assert isinstance(statement.content[3],
                          fparser.one.statements.Assignment)
    gen = str(tree)
    print(gen)
    assert "SELECT CASE ( iflag )" in gen


@pytest.mark.xfail(reason="fparser does not work with named select statements")
def test_named_select_case():
    '''Test that fparser correctly recognises a named select case'''
    from fparser import api
    source_str = '''
    subroutine foo
    integer :: iflag = 1
    real    :: aval = 0.0
    incase: select case(iflag)
    case(1) incase
      aval = 1.0
    case default   incase
      aval = 0.0
    end select  incase
    end subroutine foo
    '''
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    select_list = []
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectCase):
            select_list.append(statement)
    assert len(select_list) == 2
    for statement in select_list:
        assert isinstance(statement, fparser.one.block_statements.SelectCase)
        assert isinstance(statement.content[0], fparser.one.statements.Case)
        assert isinstance(statement.content[2], fparser.one.statements.Case)
        assert isinstance(statement.content[3],
                          fparser.one.statements.Assignment)
    gen = str(tree)
    print(gen)
    assert "incase: SELECT CASE ( iflag )" in gen


def test_select_case_brackets():
    '''Test that fparser correctly parses a select case involving
    parentheses '''
    from fparser import api
    source_str = '''
    subroutine foo
    integer :: iflag(2) = 1
    real    :: aval = 0.0
    select case(iflag(1))
    case(1)
      aval = 1.0
    case default
      aval = 0.0
    end select
    end subroutine foo
    '''
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    statement = None  # Keep pylint happy
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectCase):
            break
    assert isinstance(statement, fparser.one.block_statements.SelectCase)
    assert isinstance(statement.content[0], fparser.one.statements.Case)
    assert isinstance(statement.content[2], fparser.one.statements.Case)
    assert isinstance(statement.content[3], fparser.one.statements.Assignment)
    gen = str(tree)
    print(gen)
    assert "SELECT CASE ( iflag(1) )" in gen


def test_select_type():
    '''Test that fparser correctly recognises select type'''
    from fparser import api
    source_str = '''
    subroutine foo(an_object)
    class(*) :: an_object
    real    :: aval = 0.0
    select type(an_object)
    type is (some_type)
      aval = 1.0
    class is (some_class)
      aval = 0.0
    class default
      aval = -1.0
    end select
    ! Same again but with more white space
    select   type ( an_object )
    type  is  ( some_type(i), some_other_type )
      aval = 1.0
    class  is  ( some_class(1), some_other_class )
      aval = 0.0
    class   default
      aval = -1.0
    end select
    end subroutine foo
    '''
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    select_list = []
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectType):
            select_list.append(statement)
    assert len(select_list) == 2
    for statement in select_list:
        assert isinstance(statement, fparser.one.block_statements.SelectType)
        assert isinstance(statement.content[0], fparser.one.statements.TypeIs)
        assert isinstance(statement.content[2], fparser.one.statements.ClassIs)
        assert isinstance(statement.content[3],
                          fparser.one.statements.Assignment)
        assert isinstance(statement.content[4], fparser.one.statements.ClassIs)
    gen = str(tree)
    print(gen)
    assert "SELECT TYPE ( an_object )" in gen
    assert "TYPE IS ( some_type )" in gen
    assert "TYPE IS ( some_type(i), some_other_type )" in gen
    assert "CLASS IS ( some_class )" in gen
    assert "CLASS DEFAULT" in gen
    assert "CLASS IS ( some_class(1), some_other_class )" in gen


def test_type_is_process_item(monkeypatch, capsys):
    ''' Test error condition raised in TypeIs.process_item() method '''
    from fparser import api
    source_str = '''
    subroutine foo(an_object)
    class(*) :: an_object
    real    :: aval = 0.0
    select type(an_object)
    type is (some_type)
      aval = 1.0
    class is (some_class)
      aval = 0.0
    end select
    end subroutine foo
    '''
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    statement = None  # Keeps pylint happy
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectType):
            break
    assert isinstance(statement, fparser.one.block_statements.SelectType)
    assert isinstance(statement.content[0], fparser.one.statements.TypeIs)
    typeis = statement.content[0]
    typeis.parent.name = "not_a_name"
    monkeypatch.setattr(typeis.item, "get_line",
                        lambda: "type is (blah): wrong_name")
    # Monkeypatch the typeis object so that a call to self.warning
    # (which normally results in a call to the logger) gets replaced
    # with a call to our print_wrapper() function
    monkeypatch.setattr(typeis, "warning", print_wrapper)
    typeis.process_item()
    output, _ = capsys.readouterr()
    print(output)
    assert "expected type-is-construct-name 'not_a_name' but got " in output


def test_type_is_to_fortran():
    ''' Test error condition raised in TypeIs.to_fortran() method '''
    from fparser import api
    from fparser.common.utils import ParseError
    source_str = '''
    subroutine foo(an_object)
    class(*) :: an_object
    real    :: aval = 0.0
    select type(an_object)
    type is (some_type)
      aval = 1.0
    class is (some_class)
      aval = 0.0
    end select
    end subroutine foo
    '''
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    statement = None  # Keeps pylint happy
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectType):
            break
    assert isinstance(statement, fparser.one.block_statements.SelectType)
    assert isinstance(statement.content[0], fparser.one.statements.TypeIs)
    typeis = statement.content[0]
    typeis.name = "some_name"
    fort = typeis.tofortran()
    assert "TYPE IS ( some_type ) some_name" in fort
    # Now break the internal state and check for expected exception
    typeis.items = None
    with pytest.raises(ParseError) as excinfo:
        _ = typeis.tofortran()
    assert "TYPE IS construct must have arguments" in str(excinfo.value)


def test_class_is_process_item(monkeypatch, capsys):
    ''' Test error condition raised in ClassIs.process_item() method '''
    from fparser import api
    source_str = '''
    subroutine foo(an_object)
    class(*) :: an_object
    real    :: aval = 0.0
    select type(an_object)
    class is (some_class)
      aval = 0.0
    class default
      aval = -1.0
    end select
    end subroutine foo
    '''
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    statement = None  # Keeps pylint happy
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectType):
            break
    assert isinstance(statement, fparser.one.block_statements.SelectType)
    assert isinstance(statement.content[0], fparser.one.statements.ClassIs)
    clsis = statement.content[0]
    clsis.parent.name = "not_a_name"
    monkeypatch.setattr(clsis.item, "get_line",
                        lambda: "class is (blah): wrong_name")
    # Monkeypatch the typeis object so that a call to self.warning
    # (which normally results in a call to the logger) gets replaced
    # with a call to our print_wrapper() function
    monkeypatch.setattr(clsis, "warning", print_wrapper)
    clsis.process_item()
    output, _ = capsys.readouterr()
    print(output)
    assert "expected class-construct-name 'not_a_name' but got " in output


def test_class_is_to_fortran():
    ''' Test ClassIs.to_fortran() method '''
    from fparser import api
    source_str = '''
    subroutine foo(an_object)
    class(*) :: an_object
    real    :: aval = 0.0
    select type(an_object)
    class is (some_class)
      aval = 0.0
    end select
    end subroutine foo
    '''
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    statement = None  # Keeps pylint happy
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectType):
            break
    assert isinstance(statement, fparser.one.block_statements.SelectType)
    assert isinstance(statement.content[0], fparser.one.statements.ClassIs)
    clsis = statement.content[0]
    clsis.name = "some_name"
    fort = clsis.tofortran()
    assert "CLASS IS ( some_class ) some_name" in fort



# Modified work Copyright (c) 2017 Science and Technology Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
from fparser import api

def test_reproduce_issue():
    source_str = '''
    subroutine foo
    implicit none
    character a*8
    real*4 r
    character*2 b*4
    end subroutine foo
    '''
    tree = api.parse(source_str, isfree=True, isstrict=False)
    foo = tree.content[0]
    a = foo.get_variable('a')
    b = foo.get_variable('b')
    r = foo.get_variable('r')
    assert a.typedecl.__class__.__name__=='Character'
    assert a.length == '8',repr(a.length)
    assert a.name =='a'
    assert a.typedecl.selector==('','')
    assert r.typedecl.__class__.__name__=='Real'
    assert r.typedecl.selector==('4','')

    assert b.typedecl.__class__.__name__=='Character'
    assert b.length == '4',repr(a.length)
    assert b.name =='b'
    assert b.typedecl.selector==('2','')

def test_recursive():
    source_str = '''
    recursive subroutine foo
    end subroutine foo
    '''
    tree = api.parse(source_str, isfree=True, isstrict=False)
    foo = tree.content[0]
    assert foo.is_recursive()



# Copyright (c) 2019 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Test Fortran Include Statement: This file tests the parsing of an
include statement. Whilst include is not part of the standard Fortran
rules (the include should include code as the code is being parsed)
there are cases where users might like to keep the include statement
in the Fortran parse tree and output it again.

'''

import pytest
from fparser.api import get_reader
from fparser.two.Fortran2003 import Include_Stmt, InternalError
from fparser.two.utils import NoMatchError


def test_include_stmt(f2003_create):
    '''Check that a basic include statement is parsed
    correctly. Input separately as a string and as a reader object

    '''
    def check_include(reader):
        '''Internal helper function to avoid code replication.'''
        ast = Include_Stmt(reader)
        assert "INCLUDE 'my-non-existant-file.inc'" in str(ast)
        assert repr(ast).replace("u'", "'") == \
            ("Include_Stmt(Include_Filename("
             "'my-non-existant-file.inc'))")

    line = "include 'my-non-existant-file.inc'"
    check_include(line)
    reader = get_reader(line)
    check_include(reader)


def test_spaces(f2003_create):
    '''Check that spaces are allowed before and after an include keyword
    as well as after the file string.

    '''
    line = " include 'my-non-existant-file.inc' "
    ast = Include_Stmt(line)
    assert "INCLUDE 'my-non-existant-file.inc'" in str(ast)


def test_no_space(f2003_create):
    '''Check that no space is required between the include keyword and the
    file string.

    '''
    line = "include'my-non-existant-file.inc'"
    ast = Include_Stmt(line)
    assert "INCLUDE 'my-non-existant-file.inc'" in str(ast)


def test_case(f2003_create):
    '''Check that different case is allowed for the include keyword.'''
    line = "InClUdE 'my-non-existant-file.inc'"
    ast = Include_Stmt(line)
    assert "INCLUDE 'my-non-existant-file.inc'" in str(ast)


def test_double_quotes(f2003_create):
    '''Check that double quotes are allowed for the file string.'''
    line = 'include "my-non-existant-file.inc"'
    ast = Include_Stmt(line)
    assert "INCLUDE 'my-non-existant-file.inc'" in str(ast)


def test_errors(f2003_create):
    '''Check that syntax errors produce a NoMatchError exception.'''
    for line in [None, "", "  ", "includ", "includ 'x'", "include",
                 "include ''", "include \"x'", "include 'x\"", "include 'xxx",
                 "include \"xxx", "include xxx'", "include xxx\"",
                 "include x'x'", "include 'x'x", "x include 'x'"]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Include_Stmt(line)
        assert "Include_Stmt: '{0}'".format(line) in str(excinfo.value)


def test_include_filename_error(f2003_create, monkeypatch):
    '''Check that we raise an InternalError if a return from
    Include_Filename is None or an empty string. This should never
    happen as any matching errors would cause this class to raise an
    exception.

    '''

    monkeypatch.setattr("fparser.two.Fortran2003.Include_Filename",
                        lambda file_name: None)
    line = "include ' '"
    with pytest.raises(InternalError) as excinfo:
        _ = Include_Stmt(line)
    assert ("Include_Filename should never return None or an empty "
            "name") in str(excinfo.value)
